/*
 * clamz - Command-line downloader for the Amazon.com MP3 store
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <ctype.h>

#include <curl/curl.h>

#include "clamz.h"

struct _clamz_downloader {
  const clamz_config *cfg;
  CURL *curl;
  char *filename;
  int outfd;
  clamz_track *track;
  int last_progress;
};

/* Initialize downloader state */
clamz_downloader *new_downloader(const clamz_config *cfg)
{
  clamz_downloader *dl = malloc(sizeof(clamz_downloader));
  char *cookiejar;

  if (!dl) {
    print_error("Out of memory");
    return NULL;
  }

  if (!cfg->printonly) {
    dl->curl = curl_easy_init();

    if (!dl->curl) {
      print_error("Unable to initialize curl");
      free(dl);
      return NULL;
    }

    curl_easy_setopt(dl->curl, CURLOPT_NOPROGRESS, 0L);
    curl_easy_setopt(dl->curl, CURLOPT_FAILONERROR, 1L);
    curl_easy_setopt(dl->curl, CURLOPT_FOLLOWLOCATION, 1L);
    curl_easy_setopt(dl->curl, CURLOPT_COOKIEFILE, "");
    curl_easy_setopt(dl->curl, CURLOPT_USERAGENT, PACKAGE_STRING);

    cookiejar = get_config_file_name("cookies");
    if (cookiejar) {
      curl_easy_setopt(dl->curl, CURLOPT_COOKIEJAR, cookiejar);
      free(cookiejar);
    }
  }
  else
    dl->curl = NULL;

  dl->cfg = cfg;
  dl->filename = NULL;
  dl->outfd = -1;
  dl->track = NULL;
  return dl;
}

/* Free downloader state */
void free_downloader(clamz_downloader *dl)
{
  if (dl->curl)
    curl_easy_cleanup(dl->curl);
  if (dl->filename)
    free(dl->filename);
  if (dl->outfd > -1)
    close(dl->outfd);
  free(dl);
}

/* Convert a string according to user's preferences */
static char *convert_string(const clamz_config *cfg, const char *s)
{
  const unsigned char *us = (const unsigned char*) s;
  char *converted;
  int len;
  int i, j;

  len = strlen(s) + 1;
  converted = malloc(len * sizeof(char));

  if (!converted) {
    print_error("Out of memory");
    return NULL;
  }

  i = j = 0;
  while (s[i]) {
    if (j >= len - 1) {
      len *= 2;
      converted = realloc(converted, len * sizeof(char));

      if (!converted) {
	print_error("Out of memory");
	return NULL;
      }
    }

    if (us[i] & 0x80) {
      if (cfg->allowutf8) {
	converted[j++] = s[i++];
      }
      else {
	converted[j++] = '_';
	do
	  i++;
	while (((unsigned char) s[i] & 0xc0) == 0x80);
      }
    }
    else if (s[i] == '/' || iscntrl(us[i])) {
      converted[j++] = '_';
      i++;
    }
    else if (isupper(us[i]) && !cfg->allowupper) {
      converted[j++] = tolower(s[i++]);
    }
    else if (cfg->forbid_chars && strchr(cfg->forbid_chars, s[i])) {
      converted[j++] = '_';
      i++;
    }
    else {
      converted[j++] = s[i++];
    }
  }

  converted[j] = 0;

  return converted;
}

static int var_match(const char *str, size_t n, const char *name)
{
  if (strlen(name) != n)
    return 0;
  else
    return !strncasecmp(str, name, n);
}

static int append_subst_str(char **filename, const char *format,
			    clamz_downloader *dl)
{
  const char *p, *q;
  const char *f = format;
  const char *s;
  char* converted;

  while (*f) {
    if ((p = strchr(f, '$'))) {
      if (p != f
	  && concatenate(filename, f, p - f))
	return 1;

      p++;
      if (*p == '{') {
	p++;
	q = strchr(p, '}');
	if (!q) {
	  print_error("Invalid variable reference in '%s'", format);
	  return 1;
	}
	f = q + 1;
      }
      else {
	q = p;
	while (isalnum((int) (unsigned char) *q) || *q == '_')
	  q++;
	f = q;
      }

      if (p == q)
	s = "$";
      else if (var_match(p, q - p, "title"))
	s = dl->track->title;
      else if (var_match(p, q - p, "creator"))
	s = dl->track->creator;
      else if (var_match(p, q - p, "album"))
	s = dl->track->album;
      else if (var_match(p, q - p, "tracknum")) {
	s = dl->track->trackNum;
	if (s[0] && !s[1]) {
	  if (concatenate(filename, "0", 1))
	    return 1;
	}
      }
      else if (var_match(p, q - p, "album_artist"))
	s = find_meta(dl->track->meta, TMETA_ALBUM_ARTIST);
      else if (var_match(p, q - p, "genre"))
	s = find_meta(dl->track->meta, TMETA_GENRE);
      else if (var_match(p, q - p, "discnum"))
	s = find_meta(dl->track->meta, TMETA_DISC_NUM);
      else if (var_match(p, q - p, "suffix")) {

	// FIXME: The following doesn't work.  It might be possible to
	// infer the file type from any of the following:
	// - the fact that the download script is called "log_action.mp3"
	// - the real URL, which is URL-encoded and passed as the URL
	//   argument to the said script
	// - the content type, likewise
	// - by actually starting the download and looking for the
	//   real URL to which we're redirected a/o the content type
	//   presented to us.  I don't like this option; for safety I
	//   much prefer to have the filename known and the file opened
	//   before the transfer
	//   starts.

	/*
	if (dl->track->location) {
	  if (!(p = strchr(dl->track->location, '?')))
	    p = dl->track->location + strlen(dl->track->location);
	  q = p;
	  while (p != dl->track->location
		 && p[-1] != '/'
		 && p[-1] != '.')
	    p--;

	  if (p != dl->track->location && p[-1] == '.')
	    if (concatenate(filename, p, q - p))
	      return 1;
	}
	s = NULL;
	*/

	s = "mp3";
      }
      else if (var_match(p, q - p, "asin"))
	s = find_meta(dl->track->meta, TMETA_ASIN);
      else if (var_match(p, q - p, "album_asin"))
	s = find_meta(dl->track->meta, TMETA_ALBUM_ASIN);
      else if (var_match(p, q - p, "amz_title"))
	s = dl->track->playlist->title;
      else if (var_match(p, q - p, "amz_creator"))
	s = dl->track->playlist->creator;
      else if (var_match(p, q - p, "amz_asin"))
	s = find_meta(dl->track->playlist->meta, PMETA_ASIN);
      else if (var_match(p, q - p, "amz_genre"))
	s = find_meta(dl->track->playlist->meta, PMETA_GENRE);
      else
	s = NULL;

      if (s) {
	converted = convert_string(dl->cfg, s);
	if (!converted)
	  return 1;
	if (concatenate(filename, converted, strlen(converted))) {
	  free(converted);
	  return 1;
	}
	free(converted);
      }
    }
    else {
      return concatenate(filename, f, strlen(f));
    }
  }

  return 0;
}

static int create_parents(char *filename)
{
  char *p;

  p = filename;
  while (p) {
    p = strchr(p, '/');
    if (p) {
      *p = 0;
      if (mkdir(filename, 0755) && errno != EEXIST) {
	print_error("Cannot create directory %s: %s", filename,
		    strerror(errno));
	*p = '/';
	return 1;
      }
      *p = '/';
      p++;
    }
  }

  return 0;
}


static size_t write_output(void *ptr, size_t size, size_t n, void *data)
{
  clamz_downloader *dl = data;
  int r;

  r = write(dl->outfd, ptr, size * n);
  if (r < 0) {
    print_error("Error writing to %s: %s", dl->filename, strerror(errno));
    return 0;
  }
  else
    return r;
}

static int show_progress(void *data, double dltotal, double dlnow,
			 double ultotal UNUSED, double ulnow UNUSED)
{
  clamz_downloader *dl = data;
  int progress;

  if (dl->cfg->quiet)
    return 0;

  if (dltotal > 0)
    progress = (int) (100 * dlnow / dltotal);
  else
    progress = -1;

  if (progress != dl->last_progress) {
    dl->last_progress = progress;
    print_progress(dl->track, dl->filename, progress);
  }

  return 0;
}

int download_track(clamz_downloader *dl, clamz_track *tr)
{
  int i;
  char *s;
  CURLcode err;

  if (!tr->location) {
    print_error("No URL provided for this track");
    return 2;
  }

  dl->track = tr;

  if (dl->filename)
    free(dl->filename);

  dl->filename = NULL;

  if (dl->cfg->output_dir) {
    if (append_subst_str(&dl->filename, dl->cfg->output_dir, dl))
      return 1;
    if (append_subst_str(&dl->filename, "/", dl))
      return 1;
  }

  if (dl->cfg->name_format) {
    if (append_subst_str(&dl->filename, dl->cfg->name_format, dl))
      return 1;
  }
  else {
    print_error("No output filename specified");
    return 1;
  }

  if (!access(dl->filename, F_OK)) {
    s = malloc((strlen(dl->filename) + 10) * sizeof(char));
    if (!s) {
      print_error("Out of memory");
      return 1;
    }

    i = 1;
    do {
      sprintf(s, "%s.%d", dl->filename, i);
      i++;
    } while (!access(s, F_OK));

    print_error("\"%s\" already exists; renaming new file to \"%s\"",
		dl->filename, s);

    free(dl->filename);
    dl->filename = s;
  }

  if (dl->cfg->printonly) {
    printf("  Output to \"%s\"\n", dl->filename);
    free(dl->filename);
    dl->filename = NULL;
    return 0;
  }

  if (create_parents(dl->filename)) {
    free(dl->filename);
    dl->filename = NULL;
    return 4;
  }
 
  dl->outfd = open(dl->filename, O_WRONLY | O_APPEND | O_CREAT, 0444);
  if (dl->outfd < 0) {
    print_error("Unable to open \"%s\" (%s)", dl->filename, strerror(errno));

    free(dl->filename);
    dl->filename = NULL;
    return 4;
  }

  dl->last_progress = -2;

  curl_easy_setopt(dl->curl, CURLOPT_WRITEFUNCTION, write_output);
  curl_easy_setopt(dl->curl, CURLOPT_WRITEDATA, dl);

  curl_easy_setopt(dl->curl, CURLOPT_PROGRESSFUNCTION, show_progress);
  curl_easy_setopt(dl->curl, CURLOPT_PROGRESSDATA, dl);

  curl_easy_setopt(dl->curl, CURLOPT_URL, tr->location);

  err = curl_easy_perform(dl->curl);

  if (err) {
    print_error("Error downloading file: %s", curl_easy_strerror(err));
    close(dl->outfd);
    dl->outfd = -1;
    free(dl->filename);
    dl->filename = NULL;
    return 4;
  }

  if (close(dl->outfd)) {
    print_error("Error writing to %s", dl->filename);
    dl->outfd = -1;
    free(dl->filename);
    dl->filename = NULL;
    return 4;
  }

  dl->outfd = -1;
  free(dl->filename);
  dl->filename = NULL;
  return 0;
}
